@@ -3,23 +3,33 @@
 - raises HTTPErrors
 - creates REST API models
 """
+
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
+from __future__ import annotations
+
 import asyncio
 import os
+import pathlib  # noqa: TCH003
+import sys
 import typing as t
 import warnings
 from collections import defaultdict
 from datetime import datetime, timedelta
-from functools import partial
-from typing import Dict as DictType
-from typing import Optional
+from functools import partial, wraps
 
 from jupyter_client.ioloop.manager import AsyncIOLoopKernelManager
 from jupyter_client.multikernelmanager import AsyncMultiKernelManager, MultiKernelManager
 from jupyter_client.session import Session
 from jupyter_core.paths import exists
 from jupyter_core.utils import ensure_async
+from jupyter_events import EventLogger
+from jupyter_events.schema_registry import SchemaRegistryException
+
+if sys.version_info >= (3, 12):
+    from typing import override
+else:
+    from overrides import overrides as override
 from tornado import web
 from tornado.concurrent import Future
 from tornado.ioloop import IOLoop, PeriodicCallback
@@ -37,6 +47,7 @@ from traitlets import (
     validate,
 )
 
+from jupyter_server import DEFAULT_EVENTS_SCHEMA_PATH
 from jupyter_server._tz import isoformat, utcnow
 from jupyter_server.prometheus.metrics import KERNEL_CURRENTLY_RUNNING_TOTAL
 from jupyter_server.utils import ApiPath, import_item, to_os_path
@@ -59,7 +70,7 @@ class MappingKernelManager(MultiKernelManager):
 
     _kernel_connections = Dict()
 
-    _kernel_ports: DictType[str, t.List[int]] = Dict()  # type: ignore
+    _kernel_ports: dict[str, list[int]] = Dict()  # type: ignore[assignment]
 
     _culler_callback = None
 
@@ -67,10 +78,9 @@ class MappingKernelManager(MultiKernelManager):
 
     @default("root_dir")
     def _default_root_dir(self):
-        try:
-            return self.parent.root_dir
-        except AttributeError:
+        if not self.parent:
             return os.getcwd()
+        return self.parent.root_dir
 
     @validate("root_dir")
     def _update_root_dir(self, proposal):
@@ -199,10 +209,10 @@ class MappingKernelManager(MultiKernelManager):
         self._kernel_connections.pop(kernel_id, None)
         self._kernel_ports.pop(kernel_id, None)
 
-    # TODO DEC 2022: Revise the type-ignore once the signatures have been changed upstream
+    # TODO: DEC 2022: Revise the type-ignore once the signatures have been changed upstream
     # https://github.com/jupyter/jupyter_client/pull/905
     async def _async_start_kernel(  # type:ignore[override]
-        self, *, kernel_id: Optional[str] = None, path: Optional[ApiPath] = None, **kwargs: str
+        self, *, kernel_id: str | None = None, path: ApiPath | None = None, **kwargs: str
     ) -> str:
         """Start a kernel for a session and return its kernel_id.
 
@@ -227,18 +237,25 @@ class MappingKernelManager(MultiKernelManager):
                 kwargs["kernel_id"] = kernel_id
             kernel_id = await self.pinned_superclass._async_start_kernel(self, **kwargs)
             self._kernel_connections[kernel_id] = 0
-            task = asyncio.create_task(self._finish_kernel_start(kernel_id))
-            if not getattr(self, "use_pending_kernels", None):
-                await task
-            else:
-                self._pending_kernel_tasks[kernel_id] = task
+
             # add busy/activity markers:
             kernel = self.get_kernel(kernel_id)
             kernel.execution_state = "starting"  # type:ignore[attr-defined]
             kernel.reason = ""  # type:ignore[attr-defined]
             kernel.last_activity = utcnow()  # type:ignore[attr-defined]
             self.log.info("Kernel started: %s", kernel_id)
-            self.log.debug("Kernel args: %r", kwargs)
+            self.log.debug(
+                "Kernel args (excluding env): %r", {k: v for k, v in kwargs.items() if k != "env"}
+            )
+            env = kwargs.get("env", None)
+            if env and isinstance(env, dict):  # type:ignore[unreachable]
+                self.log.debug("Kernel argument 'env' passed with: %r", list(env.keys()))  # type:ignore[unreachable]
+
+            task = asyncio.create_task(self._finish_kernel_start(kernel_id))
+            if not getattr(self, "use_pending_kernels", None):
+                await task
+            else:
+                self._pending_kernel_tasks[kernel_id] = task
 
             # Increase the metric of number of kernels running
             # for the relevant kernel type by 1
@@ -332,7 +349,7 @@ class MappingKernelManager(MultiKernelManager):
         """
 
         if not self.buffer_offline_messages:
-            for _, stream in channels.items():
+            for stream in channels.values():
                 stream.close()
             return
 
@@ -369,7 +386,7 @@ class MappingKernelManager(MultiKernelManager):
         """
         self.log.debug("Getting buffer for %s", kernel_id)
         if kernel_id not in self._kernel_buffers:
-            return
+            return None
 
         buffer_info = self._kernel_buffers[kernel_id]
         if buffer_info["session_key"] == session_key:
@@ -436,7 +453,7 @@ class MappingKernelManager(MultiKernelManager):
         kernel = self.get_kernel(kernel_id)
         # return a Future that will resolve when the kernel has successfully restarted
         channel = kernel.connect_shell()
-        future: Future = Future()
+        future: Future[Any] = Future()
 
         def finish():
             """Common cleanup when restart finishes/fails for any reason."""
@@ -516,7 +533,8 @@ class MappingKernelManager(MultiKernelManager):
                 model = self.kernel_model(kernel_id)
                 kernels.append(model)
             except (web.HTTPError, KeyError):
-                pass  # Probably due to a (now) non-existent kernel, continue building the list
+                # Probably due to a (now) non-existent kernel, continue building the list
+                pass
         return kernels
 
     # override _check_kernel_id to raise 404 instead of KeyError
@@ -526,6 +544,40 @@ class MappingKernelManager(MultiKernelManager):
             raise web.HTTPError(404, "Kernel does not exist: %s" % kernel_id)
 
     # monitoring activity:
+    untracked_message_types = List(
+        trait=Unicode(),
+        config=True,
+        default_value=[
+            "comm_info_request",
+            "comm_info_reply",
+            "kernel_info_request",
+            "kernel_info_reply",
+            "shutdown_request",
+            "shutdown_reply",
+            "interrupt_request",
+            "interrupt_reply",
+            "debug_request",
+            "debug_reply",
+            "stream",
+            "display_data",
+            "update_display_data",
+            "execute_input",
+            "execute_result",
+            "error",
+            "status",
+            "clear_output",
+            "debug_event",
+            "input_request",
+            "input_reply",
+        ],
+        help="""List of kernel message types excluded from user activity tracking.
+
+        This should be a superset of the message types sent on any channel other
+        than the shell channel.""",
+    )
+
+    def track_message_type(self, message_type):
+        return message_type not in self.untracked_message_types
 
     def start_watching_activity(self, kernel_id):
         """Start watching IOPub messages on a kernel for activity.
@@ -546,15 +598,28 @@ class MappingKernelManager(MultiKernelManager):
 
         def record_activity(msg_list):
             """Record an IOPub message arriving from a kernel"""
-            self.last_kernel_activity = kernel.last_activity = utcnow()
-
             idents, fed_msg_list = session.feed_identities(msg_list)
             msg = session.deserialize(fed_msg_list, content=False)
 
             msg_type = msg["header"]["msg_type"]
+            parent_header = msg.get("parent_header")
+            parent_msg_type = None if parent_header is None else parent_header.get("msg_type")
+            if (
+                self.track_message_type(msg_type)
+                or self.track_message_type(parent_msg_type)
+                or kernel.execution_state == "busy"
+            ):
+                self.last_kernel_activity = kernel.last_activity = utcnow()
             if msg_type == "status":
                 msg = session.deserialize(fed_msg_list)
-                kernel.execution_state = msg["content"]["execution_state"]
+                execution_state = msg["content"]["execution_state"]
+                if self.track_message_type(parent_msg_type):
+                    kernel.execution_state = execution_state
+                elif kernel.execution_state == "starting" and execution_state != "starting":
+                    # We always normalize post-starting execution state to "idle"
+                    # unless we know that the status is in response to one of our
+                    # tracked message types.
+                    kernel.execution_state = "idle"
                 self.log.debug(
                     "activity on %s: %s (%s)",
                     kernel_id,
@@ -640,6 +705,9 @@ class MappingKernelManager(MultiKernelManager):
             await ensure_async(self.shutdown_kernel(kernel_id))
             return
 
+        kernel_spec_metadata = kernel.kernel_spec.metadata
+        cull_idle_timeout = kernel_spec_metadata.get("cull_idle_timeout", self.cull_idle_timeout)
+
         if hasattr(
             kernel, "last_activity"
         ):  # last_activity is monkey-patched, so ensure that has occurred
@@ -652,7 +720,7 @@ class MappingKernelManager(MultiKernelManager):
             dt_now = utcnow()
             dt_idle = dt_now - kernel.last_activity
             # Compute idle properties
-            is_idle_time = dt_idle > timedelta(seconds=self.cull_idle_timeout)
+            is_idle_time = dt_idle > timedelta(seconds=cull_idle_timeout)
             is_idle_execute = self.cull_busy or (kernel.execution_state != "busy")
             connections = self._kernel_connections.get(kernel_id, 0)
             is_idle_connected = self.cull_connected or not connections
@@ -702,6 +770,78 @@ class AsyncMappingKernelManager(MappingKernelManager, AsyncMultiKernelManager):
         self.last_kernel_activity = utcnow()
 
 
+def emit_kernel_action_event(success_msg: str = "") -> t.Callable[..., t.Any]:
+    """Decorate kernel action methods to
+    begin emitting jupyter kernel action events.
+
+    Parameters
+    ----------
+    success_msg: str
+        A formattable string that's passed to the message field of
+        the emitted event when the action succeeds. You can include
+        the kernel_id, kernel_name, or action in the message using
+        a formatted string argument,
+        e.g. "{kernel_id} succeeded to {action}."
+
+    error_msg: str
+        A formattable string that's passed to the message field of
+        the emitted event when the action fails. You can include
+        the kernel_id, kernel_name, or action in the message using
+        a formatted string argument,
+        e.g. "{kernel_id} failed to {action}."
+    """
+
+    def wrap_method(method):
+        @wraps(method)
+        async def wrapped_method(self, *args, **kwargs):
+            """"""
+            # Get the method name from the
+            action = method.__name__.replace("_kernel", "")
+            # If the method succeeds, emit a success event.
+            try:
+                out = await method(self, *args, **kwargs)
+                data = {
+                    "kernel_name": self.kernel_name,
+                    "action": action,
+                    "status": "success",
+                    "msg": success_msg.format(
+                        kernel_id=self.kernel_id, kernel_name=self.kernel_name, action=action
+                    ),
+                }
+                if self.kernel_id:
+                    data["kernel_id"] = self.kernel_id
+                self.emit(
+                    schema_id="https://events.jupyter.org/jupyter_server/kernel_actions/v1",
+                    data=data,
+                )
+                return out
+            # If the method fails, emit a failed event.
+            except Exception as err:
+                data = {
+                    "kernel_name": self.kernel_name,
+                    "action": action,
+                    "status": "error",
+                    "msg": str(err),
+                }
+                if self.kernel_id:
+                    data["kernel_id"] = self.kernel_id
+                # If the exception is an HTTPError (usually via a gateway request)
+                # log the status_code and HTTPError log_message.
+                if isinstance(err, web.HTTPError):
+                    msg = err.log_message or ""
+                    data["status_code"] = err.status_code
+                    data["msg"] = msg
+                self.emit(
+                    schema_id="https://events.jupyter.org/jupyter_server/kernel_actions/v1",
+                    data=data,
+                )
+                raise err
+
+        return wrapped_method
+
+    return wrap_method
+
+
 class ServerKernelManager(AsyncIOLoopKernelManager):
     """A server-specific kernel manager."""
 
@@ -710,4 +850,79 @@ class ServerKernelManager(AsyncIOLoopKernelManager):
         None, allow_none=True, help="The current execution state of the kernel"
     )
     reason = Unicode("", help="The reason for the last failure against the kernel")
+
     last_activity = Instance(datetime, help="The last activity on the kernel")
+
+    # A list of pathlib objects, each pointing at an event
+    # schema to register with this kernel manager's eventlogger.
+    # This trait should not be overridden.
+    @property
+    def core_event_schema_paths(self) -> list[pathlib.Path]:
+        return [DEFAULT_EVENTS_SCHEMA_PATH / "kernel_actions" / "v1.yaml"]
+
+    # This trait is intended for subclasses to override and define
+    # custom event schemas.
+    extra_event_schema_paths: List[str] = List(
+        default_value=[],
+        help="""
+        A list of pathlib.Path objects pointing at to register with
+        the kernel manager's eventlogger.
+        """,
+    ).tag(config=True)
+
+    event_logger = Instance(EventLogger)
+
+    @default("event_logger")
+    def _default_event_logger(self):
+        """Initialize the logger and ensure all required events are present."""
+        if (
+            self.parent is not None
+            and self.parent.parent is not None
+            and hasattr(self.parent.parent, "event_logger")
+        ):
+            logger = self.parent.parent.event_logger
+        else:
+            # If parent does not have an event logger, create one.
+            logger = EventLogger()
+        # Ensure that all the expected schemas are registered. If not, register them.
+        schemas = self.core_event_schema_paths + self.extra_event_schema_paths
+        for schema_path in schemas:
+            # Try registering the event.
+            try:
+                logger.register_event_schema(schema_path)
+            # Pass if it already exists.
+            except SchemaRegistryException:
+                pass
+        return logger
+
+    def emit(self, schema_id, data):
+        """Emit an event from the kernel manager."""
+        self.event_logger.emit(schema_id=schema_id, data=data)
+
+    @override
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was started.",
+    )
+    async def start_kernel(self, *args, **kwargs):
+        return await super().start_kernel(*args, **kwargs)
+
+    @override
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was shutdown.",
+    )
+    async def shutdown_kernel(self, *args, **kwargs):
+        return await super().shutdown_kernel(*args, **kwargs)
+
+    @override
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was restarted.",
+    )
+    async def restart_kernel(self, *args, **kwargs):
+        return await super().restart_kernel(*args, **kwargs)
+
+    @override
+    @emit_kernel_action_event(
+        success_msg="Kernel {kernel_id} was interrupted.",
+    )
+    async def interrupt_kernel(self, *args, **kwargs):
+        return await super().interrupt_kernel(*args, **kwargs)
